--[[
Copyright 2016, CZ.NIC z.s.p.o. (http://www.nic.cz/)

This file is part of NUCI configuration server.

NUCI is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NUCI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NUCI.  If not, see <http://www.gnu.org/licenses/>.
]]

require("cert");
require("datastore");
require("nutils");
require("uci");

local datastore = datastore('openvpn-client.yin');

function get_wan_ip()
	local ecode, stdout, stderr = run_command(
		nil, '/bin/ubus', '-S', 'call', 'network.interface.wan', 'status');
	if ecode == 0 then
		-- note that this doesn't have to be so accurate
		local ip = stdout:match('"ipv4%-address":%[[^%]]*"address":"(%d+%.%d+%.%d+%.%d+)"');
		if ip then
			return ip;
		end
	end
	return "<insert-ip-or-hostname>";
end

function cond_true_false(condition, if_true, if_false)
	if condition then return if_true else return if_false end
end

function build_config(settings)
	local text = [[
##############################################
# Openvpn client configuration generated by  #
# router Turris based on Sample client-side  #
# OpenVPN 2.0 config file                    #
#                                            #
# This configuration can be used only on     #
# a single client.                           #
#                                            #
#                                            #
# On Windows, you might want to rename this  #
# file so it has a .ovpn extension           #
##############################################

client

# Use the same setting as you are using on
# the server.
# On most systems, the VPN will not function
# unless you partially or fully disable
# the firewall for the TUN/TAP interface.
%s

# Windows needs the TAP-Win32 adapter name
# from the Network Connections panel
# if you have more than one.  On XP SP2,
# you may need to disable the firewall
# for the TAP adapter.
;dev-node MyTap

%s

# The hostname/IP and port of the server.
# You can have multiple remote entries
# to load balance between the servers.
;remote my-server-1 1194
;remote my-server-2 1194
%s

# Choose a random host from the remote
# list for load-balancing.  Otherwise
# try hosts in the order specified.
;remote-random

# Keep trying indefinitely to resolve the
# host name of the OpenVPN server.  Very useful
# on machines which are not permanently connected
# to the internet such as laptops.
resolv-retry infinite

# Most clients don't need to bind to
# a specific local port number.
nobind

# Downgrade privileges after initialization (non-Windows only)
;user nobody
;group nobody

# Try to preserve some state across restarts.
persist-key
persist-tun

# If you are connecting through an
# HTTP proxy to reach the actual OpenVPN
# server, put the proxy server/IP and
# port number here.  See the man page
# if your proxy server requires
# authentication.
;http-proxy-retry # retry on connection failures
;http-proxy [proxy server] [proxy port #]

# Wireless networks often produce a lot
# of duplicate packets.  Set this flag
# to silence duplicate packet warnings.
mute-replay-warnings

%s
%s
%s

remote-cert-tls server

%s

%s

%s

# Set log file verbosity.
verb 3

# Silence repeating messages
;mute 20

# To enable to process DNS push request from the server on linux machines (non systemd-resolved)
# note that you might need to have resolvconf program installed
;script-security 2
;up /etc/openvpn/update-resolv-conf
;down /etc/openvpn/update-resolv-conf

# To enable to process DNS push request from the server on linux machines (systemd-resolved)
# see https://github.com/jonathanio/update-systemd-resolved
;script-security 2
;setenv PATH /usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin
;up /etc/openvpn/update-systemd-resolved
;down /etc/openvpn/update-systemd-resolved
;down-pre

]];
	return text:format(
		'dev ' .. settings.dev,
		'proto ' .. settings.proto,
		'remote ' .. settings.remote,  -- ip and port
		'<ca>\n' .. settings.ca .. '</ca>',
		'<cert>\n' .. settings.cert .. '</cert>',
		'<key>\n' .. settings.key .. '</key>',
		cond_true_false(settings.tls_auth, 'key-direction 1\n<tls-auth>\n' .. (settings.tls_auth or '') .. '</tls_auth>', ''),
		cond_true_false(settings.cipher, 'cipher ' .. (settings.cipher or ''), ''),
		cond_true_false(settings.comp_lzo, 'comp-lzo', '')
	)
end

function datastore:user_rpc(rpc, data)
	local xml = xmlwrap.read_memory(data);
	local root = xml:root();

	if rpc == 'download-config' then
		-- get certificate name
		local cert_serial_node = find_node_name_ns(root, 'cert-serial', self.model_ns);
		if not cert_serial_node then
			return nil, {
				msg = "Missing the <cert-serial> parameter.",
				app_tag = 'data-missing',
				info_badelem = 'cert-serial',
				info_badns = self.model_ns
			}
		end
		local cert_serial = cert_serial_node:text();
		-- validate serial
		if not cert_serial or not cert_serial:match('^%x+$') then
			return nil, {
			msg = "Invalid cert-serial",
			app_tag = 'invalid-value',
			info_badelem = 'cert-serial',
			info_badns = self.model_ns
			}
		end

		-- get config name
		local config_name_node = find_node_name_ns(root, 'config-name', self.model_ns);
		if not config_name_node then
			return nil, {
				msg = "Missing the <config-name> parameter.",
				app_tag = 'data-missing',
				info_badelem = 'config-name',
				info_badns = self.model_ns
			}
		end
		local config_name = config_name_node:text();

		-- try to obtain the server address
		local server_address_node = find_node_name_ns(root, 'server-address', self.model_ns);
		local server_address = server_address_node and server_address_node:text();

		local settings = {};
		-- read uci
		cursor = get_uci_cursor();
		uci_data = cursor:get_all('openvpn', config_name);
		if not uci_data then
			return nil, "Server configuration is missing. (" .. config_name .. ")"
		end
		settings.dev = uci_data.dev;
		settings.proto = uci_data.proto;
		settings.ca_path = uci_data.ca;
		settings.tls_auth_path = uci_data.tls_auth;
		settings.cipher = uci_data.cipher;
		settings.comp_lzo = uci_data.comp_lzo;
		settings.port = uci_data.port;
		reset_uci_cursor();

		-- check settings
		if not settings.dev or not settings.proto or not settings.port then
			return nil, "Openvpn server is not well configured."
		end

		-- try to get server ip
		settings.remote = (server_address or get_wan_ip(wan_device)) .. " " .. settings.port;

		-- read ca
		local ca_content, err = file_content(settings.ca_path);
		if not ca_content then
			return nil, err;
		end
		settings.ca = ca_content;

		-- read client cert
		local cert_content, err = file_content(file_dirname(settings.ca_path) .. "/" .. cert_serial.. ".crt");
		if not cert_content then
			return nil, err;
		end
		settings.cert = cert_content;

		-- read key
		local key_content, err = file_content(file_dirname(settings.ca_path) .. "/" .. cert_serial .. ".key");
		if not key_content then
			return nil, err;
		end
		settings.key = key_content;

		-- read tls_auth if set
		if settings.tls_auth_path then
			local tls_auth_content, err = file_content(settings.tls_auth_path);
			if not tls_auth_content then
				return nil, err;
			end
			settings.tls_auth = tls_auth_content;
		end

		-- prepare output
		local new_xml = xmlwrap.new_xml_doc(self.model_name, self.model_ns);
		local new_root = new_xml:root();
		new_root:add_child('configuration'):set_text(build_config(settings));
		return new_xml:strdump();
	end
end

register_datastore_provider(datastore)
